﻿Add-PSSnapin microsoft.sharepoint.powershell
Import-Module .\SP2013_GlobalFunctions.psm1 -AsCustomObject -Force -DisableNameChecking

#region Core_Farm_Variables

##################################
# CORE FARM VARIABLES SECTION    #
##################################

# Database Information
$configDatabaseServer     = "SQLCL01-AG01-L01"   #Can be Edited
$configDatabaseName       = "CPS_SP2013_CONFIG"  #Can be Edited
$centralAdminDatabaseName = "CPS_SP2013_CENTRAL_ADMIN"  #Can be Edited

# Farm Credentials
$farmAccount      = "richsp\svc_spFarm"   #Can be Edited
$farmAccountPwd   = $null        # leave null or empty to be prompted for account credentials
$farmPassphrase   = 'LS1setup!' # use single quotes here avoid issues with having a $ in the password
$centralAdminPort = "443"

#endregion



#region Begin_farm_creation_process

################################
# CREATE SERVER FARM           #
################################

$secureFarmPassphrase   = ConvertTo-SecureString $farmPassphrase -AsPlainText -Force
if([String]::IsNullOrEmpty($farmAccountPwd))
{
    $farmAccountCredentials = Get-Credential -UserName $farmAccount -Message "Enter the credentials for the farm account."
}
else
{
    $secureFarmAccountPwd   = ConvertTo-SecureString $farmAccountPwd -AsPlainText -Force
    $farmAccountCredentials = New-Object System.Management.Automation.PSCredential($farmAccount, $secureFarmAccountPwd)
}

if(!(Test-Credential -Credential $farmAccountCredentials))
{
    Log-Message -Message "Invalid farm credential" -Color "Red"
    return
}

Add-PSSnapin Microsoft.SharePoint.PowerShell -ErrorAction SilentlyContinue | Out-Null
Log-Message -Message ""

######################################################
#  Provision a New SharePoint 2013 Farm
######################################################

    Log-Message -Message "Creating New SharePoint 2013 Farm"

    Log-Message -Message "`tPerforming Configuration task 1 of 7 -  Creating the config and admin databases..."
    New-SPConfigurationDatabase -DatabaseServer $configDatabaseServer -DatabaseName $configDatabaseName -FarmCredentials $farmAccountCredentials -Passphrase $secureFarmPassphrase -AdministrationContentDatabaseName  $centralAdminDatabaseName     
    if(!$?) {return}

    Remove-PSSnapin Microsoft.SharePoint.PowerShell
    Add-PSSnapin Microsoft.SharePoint.PowerShell -ErrorAction SilentlyContinue | Out-Null
 
    # Installs the Help site collection files for SharePoint 2013 Products in the current farm.
    Log-Message -Message "`tPerforming Configuration task 2 of 7 -  Installing help collections..."   
    Install-SPHelpCollection -All

    # Enforces resource security on the local server
    Log-Message -Message "`tPerforming Configuration task 3 of 7 -  Securing the SharePoint resources..."   
    Initialize-SPResourceSecurity

    # Installs and optionally provisions services on a farm
    Log-Message -Message "`tPerforming Configuration task 4 of 7 -  Registering SharePoint services..."   
    Install-SPService

    # Installs all available features.
    Log-Message -Message "`tPerforming Configuration task 5 of 7 -  Registering SharePoint features..."   
    Install-SPFeature -AllExistingFeatures | Out-Null

    # Creates a new Central Administration Web application and starts the central administration service on the local computer
    Log-Message -Message "`tPerforming Configuration task 6 of 7 -  Provisioning the SharePoint Central Administration Web Application..."   
    New-SPCentralAdministration -Port $centralAdminPort -WindowsAuthProvider "NTLM"

    #belt and brace - force central admin onto 443
    #note you must apply a certificate to central admin website to be able to browse it
    Set-SPCentralAdministration -Port 443  

    # Copies shared application data to existing Web application folders.
    Log-Message -Message "`tPerforming Configuration task 7 of 7 -  Installing the application content files...`n"   
    Install-SPApplicationContent

    # Start OWSTIMER
    Start-Service SPTimerv4

#endregion

#region Create_Managed_Accounts_Required   #Can be Edited

$managedAccounts = @(
     "svc_spservice"     #Can be Edited
    ,"svc_spcontent"        #Can be Edited
)


foreach($account in $managedAccounts)
{
    $managedAccount = Get-SPManagedAccount -Identity $account -ErrorAction SilentlyContinue
 
    if(!$managedAccount)
    {
        $message = "Please enter the password for account: {0}" -f $account
        $credential = Get-Credential -Message $message -UserName $account

        $managedAccount = New-SPManagedAccount -Credential $credential        
    }
}

#endregion





