﻿#region Param

param
(
    [String]$DomainName,

    [String]$DomainNetbiosName,

    [PSCredential]$Credential,

    [String]$OrganizationName
)

#endregion

#region Decrypt

function Decrypt
{
    param
    (
        [Parameter(Mandatory)]
        [String]$Thumbprint,

        [Parameter(Mandatory)]
        [String]$Base64EncryptedValue
    )

    # Decode Base64 string
    $encryptedBytes = [System.Convert]::FromBase64String($Base64EncryptedValue)

    # Get certificate from store
    $store = new-object System.Security.Cryptography.X509Certificates.X509Store([System.Security.Cryptography.X509Certificates.StoreLocation]::LocalMachine)
    $store.open([System.Security.Cryptography.X509Certificates.OpenFlags]::ReadOnly)
    $certificate = $store.Certificates | %{if($_.thumbprint -eq $Thumbprint){$_}}
   
    # Decrypt
    $decryptedBytes = $certificate.PrivateKey.Decrypt($encryptedBytes, $false)
    $decryptedValue = [System.Text.Encoding]::UTF8.GetString($decryptedBytes)
    
    return $decryptedValue
}

if ($EncryptionCertificateThumbprint)
{
    Write-Verbose -Message "Decrypting parameters with certificate $EncryptionCertificateThumbprint..."

    $Password = Decrypt -Thumbprint $EncryptionCertificateThumbprint -Base64EncryptedValue $Password
    $SafeModeAdministratorPassword = Decrypt -Thumbprint $EncryptionCertificateThumbprint -Base64EncryptedValue $SafeModeAdministratorPassword

    Write-Verbose -Message "Successfully decrypted parameters."
}
else
{
    Write-Verbose -Message "No encryption certificate specified. Assuming cleartext parameters."
}

#endregion

#region MBX Configuration Script

if ($env:COMPUTERNAME -match 'MBX-Server\d\d' -AND (test-path C:\Windows\temp\First.txt)) 
{
    Configuration FirstExchangeMailboxServerInstall
    {

        Import-DscResource -Module xActiveDirectory, xComputerManagement, xExchange, xPendingReboot
        node $AllNodes.Where{$_.ServerRole -eq 'Exchange Mailbox Server'}.Nodename
        {
            foreach ($Component in $Node.RequiredFeatures)
            {
                WindowsFeature $Component
                {
                    Ensure = 'Present'
                    Name = $Component
                }
            }
        
            xWaitForADDomain WaitforDomain
            {
                DomainName = $DomainName
                DomainUserCredential = $Credential
                RetryCount = 720
                RetryIntervalSec = 10
            }

            xComputer DomainJoin
            {
                Name = $env:COMPUTERNAME
                DomainName = $DomainName
                Credential = $Credential
                DependsOn =  '[xWaitForADDomain]WaitforDomain'
            }
            
            xWaitForDisk DataDisk
            {
                DiskNumber = $Node.DataDisk
                RetryCount = 720
            }

            xWaitForDisk SourceDisk
            {
                DiskNumber = $Node.SourceDisk
                RetryCount = 720
            }

            xDisk DataDisk
            {
                DiskNumber = $Node.DataDisk
                DriveLetter = $Node.Data
                DependsOn = '[xWaitForDisk]DataDisk'
            }

            xDisk SourceDisk
            {
                DiskNumber = $Node.SourceDisk
                DriveLetter = $Node.Source
                DependsOn = '[xWaitForDisk]SourceDisk'
            }

            xPendingReboot Reboot
            {
                Name = "BeforeAppInstalls"
                DependsOn = '[xDisk]DataDisk','[xDisk]SourceDisk','[xComputer]DomainJoin','[WindowsFeature]Server-Media-Foundation','[WindowsFeature]RSAT-AD-PowerShell'
            }
        
            Package UCMA
            {
                Ensure = 'Present'
                Name = 'Unified Communications Managed API 4.0 Runtime'
                Credential = $Credential
                Path = (join-path -path $Node.Source -childpath ':\Install\UCMA\UCMARunTimeSetup.exe')
                ProductId = '{ED98ABF5-B6BF-47ED-92AB-1CDCAB964447}'
                Arguments = '/passive'
                DependsOn = '[xPendingReboot]Reboot'
            }

            xExchangeSchema ExchangeSchema
            {
                SchemaVersion = '15292'
                Ensure = 'Present'
                Organization = $Node.OrganizationName
                SourcePath = $Node.Source
                SourceFolder = ':\Install\Exchange\'
            }

            xExchangeSetup Exchange
            {
                SourcePath = $Node.Source
                SourceFolder = ':\Install\Exchange\'
                Arguments = '/mode:Install /OrganizationName:' + $Node.OrganizationName + ' /role:Mailbox,ClientAccess /Iacceptexchangeserverlicenseterms /MDBName:' + $Node.NodeName + 'Default /DbFilePath:' + $Node.Data + ':\DB\' + $Node.NodeName + 'Default.edb /LogFolderPath:' + $Node.Data + ':\Logs /TargetDir:' + $Node.Data + ':\Exchange'
                Credential = $Credential
                DependsOn = '[xDeployPrerequisiteSetup]UCMA'
            }

            LocalConfigurationManager
            {
                CertificateId = $node.Thumbprint
                ConfigurationMode = 'ApplyandAutoCorrect'
                RebootNodeIfNeeded = 'True'
            }
        }
    }

    $ConfigData = @{
        AllNodes = @(
            @{
                Nodename = $env:COMPUTERNAME
                ServerRole = 'Exchange Mailbox Server'
                DomainName = $DomainName
                SourceDisk = 2
                DataDisk = 1
                Data = 'F'
                Source = 'G'
                OrganizationName = $OrganizationName
                PSDscAllowPlainTextPassword = $true
                Credential = $Credential
                RequiredFeatures = 'AS-HTTP-Activation', 'Desktop-Experience', 'NET-Framework-45-Features', 'RPC-over-HTTP-proxy', 'RSAT-AD-PowerShell', 'RSAT-ADDS-Tools', 'RSAT-Clustering', 'RSAT-Clustering-CmdInterface', 'RSAT-Clustering-Mgmt', 'RSAT-Clustering-PowerShell', 'Server-Media-Foundation', 'Web-Asp-Net45', 'Web-Basic-Auth', 'Web-Client-Auth', 'Web-Digest-Auth', 'Web-Dir-Browsing', 'Web-Dyn-Compression', 'Web-Http-Errors', 'Web-Http-Logging', 'Web-Http-Redirect', 'Web-Http-Tracing', 'Web-ISAPI-Ext', 'Web-ISAPI-Filter', 'Web-Lgcy-Mgmt-Console', 'Web-Metabase', 'Web-Mgmt-Console', 'Web-Mgmt-Service', 'Web-Net-Ext45', 'WAS-Process-Model', 'Web-Request-Monitor', 'Web-Server', 'Web-Stat-Compression', 'Web-Static-Content', 'Web-Windows-Auth', 'Web-WMI', 'Windows-Identity-Foundation'
            }
        )
    }

FirstExchangeMailboxServerInstall -ConfigurationData $configData -OutputPath $PSScriptRoot
}

if ($env:COMPUTERNAME -match 'MBX-Server\d\d' -AND (test-path C:\Windows\temp\First.txt) -eq $false) 
{
    Configuration ExchangeMailboxServerInstall
    {

        Import-DscResource -Module xActiveDirectory, xComputerManagement, xDeploy, xPendingReboot
        node $AllNodes.Where{$_.ServerRole -eq 'Exchange Mailbox Server'}.Nodename
        {
            foreach ($Component in $Node.RequiredFeatures)
            {
                WindowsFeature $Component
                {
                    Ensure = 'Present'
                    Name = $Component
                }
            }
        
            xWaitForADDomain WaitforDomain
            {
                DomainName = $DomainName
                DomainUserCredential = $Credential
                RetryCount = 720
                RetryIntervalSec = 10
            }

            xComputer DomainJoin
            {
                Name = $env:COMPUTERNAME
                DomainName = $DomainName
                Credential = $Credential
                DependsOn =  '[xWaitForADDomain]WaitforDomain'
            }
            
            xWaitForDisk DataDisk
            {
                DiskNumber = $Node.DataDisk
                RetryCount = 720
            }

            xWaitForDisk SourceDisk
            {
                DiskNumber = $Node.SourceDisk
                RetryCount = 720
            }

            xDisk DataDisk
            {
                DiskNumber = $Node.DataDisk
                DriveLetter = $Node.Data
                DependsOn = '[xWaitForDisk]DataDisk'
            }

            xDisk SourceDisk
            {
                DiskNumber = $Node.SourceDisk
                DriveLetter = $Node.Source
                DependsOn = '[xWaitForDisk]SourceDisk'
            }

            xPendingReboot Reboot
            {
                Name = "BeforeAppInstalls"
                DependsOn = '[xDisk]DataDisk','[xDisk]SourceDisk','[xComputer]DomainJoin','[WindowsFeature]Server-Media-Foundation','[WindowsFeature]RSAT-AD-PowerShell'
            }
        
            xDeployPrerequisiteSetup UCMA
            {
                Name = 'Unified Communications Managed API 4.0 Runtime'
                Credential = $Credential
                SourcePath = $Node.Source
                SourceFolder = ':\Install\UCMA\UCMARunTimeSetup.exe'
                DependsOn = '[xPendingReboot]Reboot'
            }

            xWaitforExchangeSchema ExchangeSchema
            {
                SchemaVersion = '15292'
            }

            xExchangeSetup Exchange
            {
                SourcePath = $Node.Source
                SourceFolder = ':\Install\Exchange\Setup.exe'
                Arguments = '/mode:Install /OrganizationName:' + $Node.OrganizationName + ' /role:Mailbox,ClientAccess /Iacceptexchangeserverlicenseterms /MDBName:' + $Node.NodeName + 'Default /DbFilePath:' + $Node.Data + ':\DB\' + $Node.NodeName + 'Default.edb /LogFolderPath:' + $Node.Data + ':\Logs /TargetDir:' + $Node.Data + ':\Exchange'
                Credential = $Credential
                DependsOn = '[xDeployPrerequisiteSetup]UCMA'
            }

            LocalConfigurationManager
            {
                CertificateId = $node.Thumbprint
                ConfigurationMode = 'ApplyandAutoCorrect'
                RebootNodeIfNeeded = 'True'
            }
        }
    }

    $ConfigData = @{
        AllNodes = @(
            @{
                Nodename = $env:COMPUTERNAME
                ServerRole = 'Exchange Mailbox Server'
                DomainName = $DomainName
                SourceDisk = 2
                DataDisk = 1
                Data = 'F'
                Source = 'G'
                OrganizationName = $OrganizationName
                PSDscAllowPlainTextPassword = $true
                Credential = $Credential
                RequiredFeatures = 'AS-HTTP-Activation', 'Desktop-Experience', 'NET-Framework-45-Features', 'RPC-over-HTTP-proxy', 'RSAT-AD-PowerShell', 'RSAT-ADDS-Tools', 'RSAT-Clustering', 'RSAT-Clustering-CmdInterface', 'RSAT-Clustering-Mgmt', 'RSAT-Clustering-PowerShell', 'Server-Media-Foundation', 'Web-Asp-Net45', 'Web-Basic-Auth', 'Web-Client-Auth', 'Web-Digest-Auth', 'Web-Dir-Browsing', 'Web-Dyn-Compression', 'Web-Http-Errors', 'Web-Http-Logging', 'Web-Http-Redirect', 'Web-Http-Tracing', 'Web-ISAPI-Ext', 'Web-ISAPI-Filter', 'Web-Lgcy-Mgmt-Console', 'Web-Metabase', 'Web-Mgmt-Console', 'Web-Mgmt-Service', 'Web-Net-Ext45', 'WAS-Process-Model', 'Web-Request-Monitor', 'Web-Server', 'Web-Stat-Compression', 'Web-Static-Content', 'Web-Windows-Auth', 'Web-WMI', 'Windows-Identity-Foundation'
            }
        )
    }

ExchangeMailboxServerInstall -ConfigurationData $configData -OutputPath $PSScriptRoot
}

#endregion

#region CAS Configuration Script

if ($env:COMPUTERNAME -match 'CAS-Server\d\d') 
{
    Configuration ExchangeClientAccessServerInstall
    {
        Import-DscResource -Module xActiveDirectory, xComputerManagement, xDeploy, xPendingReboot

        node $AllNodes.Where{$_.ServerRole -eq 'Exchange Client Access Server'}.Nodename
        {

            foreach ($Component in $Node.RequiredFeatures)
            {
                WindowsFeature $Component
                {
                    Ensure = 'Present'
                    Name = $Component
                }
            }
        
            xWaitForADDomain WaitforDomain
            {
                DomainName = $DomainName
                DomainUserCredential = $Credential
                RetryCount = 720
                RetryIntervalSec = 10
            }

            xComputer DomainJoin
            {
                Name = $env:COMPUTERNAME
                DomainName = $DomainName
                Credential = $Credential
                DependsOn =  '[xWaitForADDomain]WaitforDomain'
            }

            xWaitForDisk DataDisk
            {
                DiskNumber = $Node.DataDisk
                RetryCount = 720
            }

            xWaitForDisk SourceDisk
            {
                DiskNumber = $Node.SourceDisk
                RetryCount = 720
            }

            xDisk DataDisk
            {
                DiskNumber = $Node.DataDisk
                DriveLetter = $Node.Data
                DependsOn = '[xWaitForDisk]DataDisk'
            }

            xDisk SourceDisk
            {
                DiskNumber = $Node.SourceDisk
                DriveLetter = $Node.Source
                DependsOn = '[xWaitForDisk]SourceDisk'
            }

            xPendingReboot Reboot
            {
                Name = "BeforeAppInstalls"
                DependsOn = '[xDisk]DataDisk','[xDisk]SourceDisk','[xComputer]DomainJoin','[WindowsFeature]Server-Media-Foundation','[WindowsFeature]RSAT-AD-PowerShell'
            }
        
            xDeployPrerequisiteSetup UCMA
            {
                Name = 'Unified Communications Managed API 4.0 Runtime'
                Credential = $Credential
                SourcePath = $Node.Source
                SourceFolder = ':\Install\UCMA\UCMARunTimeSetup.exe'
                DependsOn = '[xPendingReboot]Reboot'
            }   

            xWaitforExchangeSchema ExchangeSchema
            {
                SchemaVersion = '15292'
            }

            xExchangeSetup Exchange
            {
                SourcePath = $Node.Source
                SourceFolder = ':\Install\Exchange\Setup.exe'
                Arguments = '/mode:Install /OrganizationName:' + $Node.OrganizationName + ' /role:ClientAccess /Iacceptexchangeserverlicenseterms /TargetDir:' + $Node.Data + ':\Exchange'
                Credential = $Credential
                DependsOn = '[xDeployPrerequisiteSetup]UCMA'
            }

            LocalConfigurationManager
            {
                CertificateId = $node.Thumbprint
                ConfigurationMode = 'ApplyandAutoCorrect'
                RebootNodeIfNeeded = 'True'
            }
        }
    }

    $ConfigData = @{
        AllNodes = @(
            @{
                Nodename = $env:COMPUTERNAME
                ServerRole = 'Exchange Client Access Server'
                DomainName = $DomainName
                SourceDisk = 2
                DataDisk = 1
                Data = 'F'
                Source = 'G'
                OrganizationName = $OrganizationName
                PSDscAllowPlainTextPassword = $true
                Credential = $Credential
                RequiredFeatures = 'AS-HTTP-Activation', 'Desktop-Experience', 'NET-Framework-45-Features', 'RPC-over-HTTP-proxy', 'RSAT-AD-PowerShell', 'RSAT-ADDS-Tools', 'RSAT-Clustering', 'RSAT-Clustering-CmdInterface', 'RSAT-Clustering-Mgmt', 'RSAT-Clustering-PowerShell', 'Server-Media-Foundation', 'Web-Asp-Net45', 'Web-Basic-Auth', 'Web-Client-Auth', 'Web-Digest-Auth', 'Web-Dir-Browsing', 'Web-Dyn-Compression', 'Web-Http-Errors', 'Web-Http-Logging', 'Web-Http-Redirect', 'Web-Http-Tracing', 'Web-ISAPI-Ext', 'Web-ISAPI-Filter', 'Web-Lgcy-Mgmt-Console', 'Web-Metabase', 'Web-Mgmt-Console', 'Web-Mgmt-Service', 'Web-Net-Ext45', 'WAS-Process-Model', 'Web-Request-Monitor', 'Web-Server', 'Web-Stat-Compression', 'Web-Static-Content', 'Web-Windows-Auth', 'Web-WMI', 'Windows-Identity-Foundation'
            }
        )
    }

ExchangeClientAccessServerInstall -ConfigurationData $configData -OutputPath $PSScriptRoot

}
#endregion

#region Additional Config Data
    
if ($EncryptionCertificateThumbprint)
{
    $certificate = dir Cert:\LocalMachine\My\$EncryptionCertificateThumbprint
    $certificatePath = Join-Path -path $PSScriptRoot -childPath "EncryptionCertificate.cer"
    Export-Certificate -Cert $certificate -FilePath $certificatePath | Out-Null
    $configData = @{
        AllNodes = @(
            @{
                Nodename = "*"
                CertificateFile = $certificatePath
                Thumbprint = $EncryptionCertificateThumbprint
                PSDscAllowPlainTextPassword = $false
            }
        )
    }
}

#endregion

#region Apply MOF
winrm quickconfig -quiet

Set-DscLocalConfigurationManager -ComputerName $env:COMPUTERNAME -Path $PSScriptRoot -Verbose
Start-DscConfiguration -ComputerName $env:COMPUTERNAME -Path $PSScriptRoot -Force -Verbose

#endregion