﻿
<###################################################
 #                                                 #
 #  Copyright (c) Microsoft. All rights reserved.  #
 #                                                 #
 ##################################################>

<#

.SYNOPSIS

This script is used to deploy Microsoft first party workloads in the CPS system.

.DESCRIPTION

-ProductKey parameter product key for the OS image.

The credential used to change AD passwords must belong to one of the following user groups:
    Domain Administrator

- Added functionality to deploy the PCS specific Gallery artifacts and OS Source Disk

#>

[CmdletBinding()]
    Param(
        [Parameter(Mandatory=$false)]
        [string]$Password,

        [Parameter(Mandatory=$false)]
        [string]$SourceOSVHDName,

		[Parameter(Mandatory=$false)]
        [string]$OSDiskRelease = "1.1.0.0",

        [Parameter(Mandatory=$false)]
        [bool]$UpdateArtifacts = $true,

        [Parameter(Mandatory=$false)]
        [string]$SQLInstallationDiskVHDName,

        [Parameter(Mandatory=$false)]
        [string]$ExchangeInstallationDiskVHDName,

        [Parameter(Mandatory=$false)]
        [string]$SharePointInstallationDiskVHDName,

        [Parameter(Mandatory=$false)]
        [string]$SkypeForBusinessInstallationDiskVHDName,
         

		[Parameter(Mandatory=$true)]
        [string]$VMMlibraryName,

		[Parameter(Mandatory=$false)]
        [bool]$IsCPSRack = $true,

        [Parameter(Mandatory=$false)]
        [bool]$UpdatePCSArtifacts = $false
            
    )

Function Log($message, $level = $logInfo, $toFile = $true, $toConsole = $true)
{
    $global:logFileName = ".\DeployWorkload.log"
    if(-not (Test-Path $global:logFileName))
    {
        New-Item $global:logFileName -ItemType "file" -Force | Out-Null
    }

    $date = Get-Date -Format "yyyy-MM-dd HH:mm:ss"
    if ($level -eq "Info")
    {
        $message = $date + " " + $logInfo + ": " + $message
        if ($toFile)
        {
            Add-Content $global:logFileName -Value $message
        }
        if ($toConsole)
        {
            Write-Host $message
        }
    }
    elseif ($level -eq "Warning")
    {
        $message = $date + " " + $logWarning + ": " + $message
        if ($toFile)
        {
            Add-Content $global:logFileName -Value $message
        }
        if ($toConsole)
        {
            Write-Host $message -ForegroundColor Yellow
        }
    }
    elseif ($level -eq "Error")
    {
        $message = $date + " " + $logError + ": " + $message
        if ($toFile)
        {
            Add-Content $global:logFileName -Value $message
        }
        if ($toConsole)
        {
            Write-Error $message
        }
    }
}

Function Update-TenantProvisioningAssets
{
    param
    (
        [string]$assetDescriptionType = "Environmental",
        [string]$Password = "CPSworkloads19$"
    )

    $prefix = Get-MachineNamePrefix
    $smaWSEndpoint = "https://$prefix-sma-01"

    if ($assetDescriptionType -eq "Environmental"){
        $Desc = $assetDescriptionType
        
        $ServerMap = Get-SmaVariable -WebServiceEndpoint $smaWSEndpoint -Name "CpsServerMap"
        Set-SmaVariable -Name "WAP Admin API" -Value $ServerMap.Value.KATAL.'Admin API Tier'.NLBs[0].Hosts[0].Hostname -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "Default Client Realm" -Value "http://azureservices/AdminSite" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "Default Tenant Domain" -Value "contoso.com" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "WAP Server" -Value $ServerMap.Value.KATAL.'Admin API Tier'.NLBs[0].Hosts[0].Hostname -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "Default Cloud Name" -Value "Tenant Cloud" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "ADFS Server FQDN" -Value $ServerMap.Value.KATAL.'ADFS Tier'.NLBs[0].Hosts[0].Hostname -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "WAP Tenant API FQDN" -Value $ServerMap.Value.KATAL.'Tenant API Tier'.NLBs[0].Hosts[0].Hostname -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "LoadBalancerFrontendNetwork" -Value "External" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "LoadBalancerBackendNetwork" -Value "Load Balancer" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "LoadBalancerFrontendBackEndPort" -Value "443" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "WAP Tenant API FQDN" -Value $ServerMap.Value.KATAL.'Tenant API Tier'.NLBs[0].Hosts[0].Hostname -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "RDSHCollectionName" -Value "RDSHCollection1" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "RDSHConnectionBrokerNode" -Value "RDSHConnectionBrokerNode1" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "RDSHSessionHostNode" -Value "RDSHSessionHostNode1" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "RDSHVIPNAME" -Value "RDSHVIPNAME1" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "RDSHWebAccessServerNode" -Value "RDSHWebAccessServerNode1" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "SkypeFileShareName" -Value "FileShare" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        
        
        $CPSComputerMap = Get-SmaVariable -WebServiceEndpoint $smaWSEndpoint -Name "CPSComputerMap"
        Set-SmaVariable -Name "VMM Server" -Value ($CPSComputerMap.Value.'HASCVMM-SN' + "." + $CPSComputerMap.Value.'FQDN') -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "Default VMM Logical Network" -Value "Network Virtualization" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "Gateway Service Name" -Value "$prefix-1NAT0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "Default External IP Address Pool Name" -Value "ScaleUnit1-External-IPV4Pool" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        
        $WAPEndpointMap = Get-SmaVariable -WebServiceEndpoint $smaWSEndpoint -Name "WAPEndpointMap"
        Set-SmaVariable -Name "WAP Tenant Portal Address" -Value $WAPEndpointMap.Value.'WAP MgmtSvc-TenantSite FQDN' -Description $Desc -WebServiceEndpoint $smaWSEndpoint
    }
    
    if ($assetDescriptionType -eq "Disks"){
        $Desc = $assetDescriptionType
        Set-SmaVariable -Name "DSC SQL VM Role OS Disk" -Value "Windows Server 2012 R2 Datacenter:1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC SP VM Role OS Disk" -Value "Windows Server 2012 R2 Datacenter:1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC VM Role OS Disk" -Value "Windows Server 2012 R2 Datacenter:1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        $VMRoleVHDs = @{
            "DC-Server####" = "ActiveDirectory",5120,$true
            "MBX-Server####" = "Exchange",2048000,$false
            "CAS-Server####" = "Exchange",10240,$false
            "SPFE-Server####" = "SharePoint",102400,$true
            "SPAP-Server####" = "SharePoint",102400,$true
            "SP-Server####" = "SharePoint",102400,$true
            "WN-Server####" = "Witness",2048,$true
            "SQLC-Server####" = "SqlServer",512000,$true
            "SQLF-Server####" = "SqlServer",512000,$true
            "SQLS-Server####" = "SqlServer",102400,$true
            "SkypeFE####" = "SkypeFE",40000,$true
            "SkypeShare####" = "SkypeShare",40000,$true
            "SkypeEdge####" = "SkypeEdge",40000,$true
        }
        Set-SmaVariable -Name "VMRoleVHDs" -Value $VMRoleVHDs -Description $Desc -WebServiceEndpoint $smaWSEndpoint
    }
    
    if ($assetDescriptionType -eq "VMRoles"){
        $Desc = $assetDescriptionType
        Set-SmaVariable -Name "DSC AD Gallery Item" -Value "ActiveDirectorySetup;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC CAS Gallery Item" -Value "ExchangeSetupCAS;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC Witness Gallery Item" -Value "WitnessServerSetup;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC MBX Gallery Item" -Value "ExchangeSetupMBX;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC SQLC Gallery Item" -Value "SQLServerAlwaysOnSetup;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC SQLF Gallery Item" -Value "SQLServerAlwaysOnSetup;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC SPAP Gallery Item" -Value "SharePointSetup;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC SPFE Gallery Item" -Value "SharePointSetup;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC RDSH CB Gallery Item" -Value "RDSCB;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC RDSH Session Host Gallery Item" -Value "RDSH;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC RDSH Web Gateway Gallery Item" -Value "WEBGW;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        
        Set-SmaVariable -Name "DSC Skype FE Server Gallery Item" -Value "SkypeServerFE;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC Skype Edge Server Gallery Item" -Value "SkypeForBusinessServerEdge;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "DSC Skype File Share Server Gallery Item" -Value "SkypeFileShareServer;1.1.0.0" -Description $Desc -WebServiceEndpoint $smaWSEndpoint

    }

    if ($assetDescriptionType -eq "Passwords"){
        $Desc = $assetDescriptionType
        Set-SmaVariable -Name "VMRole Default Password" -Value $Password -Description $Desc -WebServiceEndpoint $smaWSEndpoint
    }

    if ($assetDescriptionType -eq "Notifications"){
        $Desc = $assetDescriptionType
        Set-SmaVariable -Name "EmailCc" -Value "test@contoso.com" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaVariable -Name "SMTPServer" -Value "smtp.office365.com" -Description $Desc -WebServiceEndpoint $smaWSEndpoint
        Set-SmaCredential -Name "scdemobr" -Value (New-Object System.Management.Automation.PSCredential ("scdemobr@microsoft.com", (ConvertTo-SecureString "c#t!nnovatesLikeAM0F0" -AsPlainText -Force))) -Description $Desc -WebServiceEndpoint $smaWSEndpoint
    }
    
}

Function Get-MachineNamePrefix
{
    $name = hostname
    return ($name -split "-CON")[0]
}

Function Get-VMMServerConnection
{
	param ([bool]$CPSRack)
	
	$prefix = Get-MachineNamePrefix
    
	if ($CPSRack -eq $false)
     {
      Get-SCVMMServer -ComputerName "$prefix-VMM01"
     }

     else 
    
	 {
        Get-SCVMMServer -ComputerName "$prefix-HA-VMM"
     } 
}




Function Import-Runbooks
{
    $importFolder = ".\Runbooks\"
    $prefix = Get-MachineNamePrefix
    $smaWSEndpoint = "https://$prefix-sma-01"
    $Runbooks = Get-ChildItem -Path $importFolder

    foreach ($Runbook in $Runbooks) {

        $RunbookPath = $Runbook.FullName
        $RunbookName = $Runbook.Name.Split(".")[0]

        #Check to see if the Runbook Exists (for Re-Import)
        $RunbookExists = Get-SmaRunbook -Name $RunbookName -WebServiceEndpoint $smaWSEndpoint -ErrorAction SilentlyContinue

        if ($RunbookExists) {

            #Re-Import Runbooks
            Edit-SmaRunbook -Path $RunbookPath -Name $RunbookName -Overwrite:$true -WebServiceEndpoint $smaWSEndpoint

        } else {

            #Identify Tags
            if ($RunbookName -match "Remove"){
                if ($RunbookName -eq "Remove-SubscriptionResources") {$Tags = "Type:Process, Proj:Tenant Deprovisioning"} else {$Tags = "Type:Sub, Proj:Tenant Deprovisioning"}
            } elseif ($RunbookName -eq "Get-WAPSubscription"){
                $Tags = "Type:Sub, Proj:Tenant Deprovisioning"
            } elseif ($RunbookName -match "Dispatch"){
                $Tags = "Type:Working,SPF, Proj:Tenant Provisioning"
            } elseif ($RunbookName -iin "Deploy-DSCWorkloadVMRoles","Monitor-MultipleSCJobs","Monitor-VMMJobStatus","Create-TenantVirtualNetwork","Deploy-DedicatedExchangeVMRoles","Deploy-DedicatedSharePointVMRoles","Deploy-DedicatedSQLVMRoles"){
                if ($RunbookName -iin "Deploy-DSCWorkloadVMRoles","Deploy-DedicatedExchangeVMRoles","Deploy-DedicatedSharePointVMRoles","Deploy-DedicatedSQLVMRoles") {$Tags = "Type:Working,Process, Proj:Tenant Provisioning"} else {$Tags = "Type:Process, Proj:Tenant Provisioning"}
            } elseif ($RunbookName -iin "Deploy-TenantVMRole","Attach-DataVHD","Create-SubscriptionNATRules","Test-GuestVMIPPoolAllocation"){
                $Tags = "Type:Working,Sub, Proj:Tenant Provisioning"
            } else {
                $Tags = "Type:Sub, Proj:Tenant Provisioning"
            }

            #Import Runbooks
            Import-SmaRunbook -Path $RunbookPath -Tags $Tags -WebServiceEndpoint $smaWSEndpoint -Confirm:$false
        }

        #Publish Imported/Re-Imported Runbook
        Publish-SmaRunbook -Name $RunbookName -WebServiceEndpoint $smaWSEndpoint -Confirm:$false
    }
}


Function Import-GalleryItem
{
    ipmo spfadmin
    $galleryitems = Get-ChildItem -Path .\ -Recurse -Filter *.resdefpkg
    foreach($item in $galleryitems)
    {
        $gi = Import-SCSpfVMRoleGalleryItem -PackageFilePath $item.FullName -ErrorAction SilentlyContinue 
        if($gi -ne $null)
        {
            Set-SCSpfVMRoleGalleryItem -VmRoleGalleryItem $gi -PublishingStatus Published
        }
    }
}

Function New-SPFEventRegistration
{
    ipmo spfadmin
    Set-SCSPFEventRegistration -Resource "MicrosoftCompute.VMRole" -Action "Create" -Enabled -RunbookName "VMRole-Create-Dispatcher"
    Set-SCSPFEventRegistration -Resource "MicrosoftCompute.VMRole" -Action "Action_Scale" -Enabled -RunbookName "VMRole-Scale-Dispatcher"
    Set-SCSPFEventRegistration -Resource "Subscription" -Action "Create" -Enabled -RunbookName "Subscription-Create-Dispatcher"
    Set-SCSPFEventRegistration -Resource "Subscription" -Action "Delete" -Enabled -RunbookName "Subscription-Delete-Dispatcher"

}

Function Import-ResourceExtension
{
    
    Get-VMMServerConnection -CPSRack $IsCPSRack   # Determine the SCVMM Server Name based on the Rack
		
    $resexts = Get-ChildItem -Path .\ -Recurse -Filter *.resextpkg
    foreach($resext in $resexts)
    {
        if($IsCPSRack)
        {
            Import-CloudResourceExtension -ResourceExtensionPath $resext.FullName -SharePath ( Get-SCLibraryShare | ? name -eq $VMMlibraryName) -AllowUnencryptedTransfer -ErrorAction SilentlyContinue
        }
        else
        {
            Import-CloudResourceExtension -ResourceExtensionPath $resext.FullName -SharePath ( Get-SCLibraryShare | ? name -eq $VMMlibraryName) -ErrorAction SilentlyContinue
        }

       
    }
}


Function Set-DiskMetadata
{
    Param(
        [string]$DiskSharedPath,
        [string]$OperatingSystem,
        [string]$VirtualizationPlatform,
        [string]$Release,
        [string]$FamilyName,
        [string]$ProductKey

    )
    
    
	Get-VMMServerConnection -CPSRack $IsCPSRack   # Determine the SCVMM Server Name based on the Rack

    $disk = Get-SCVirtualHardDisk | ? {$_.SharePath -ieq $DiskSharedPath}
    
    if($disk -eq $null)
    {
        Log "Can not disk for shared path: $DiskSharedPath" "Error"
        return
    }

    if($OperatingSystem -eq "None")
    {
        Set-SCVirtualHardDisk -VirtualHardDisk $disk -VirtualizationPlatform $VirtualizationPlatform -OperatingSystem $OperatingSystem -Release $Release -FamilyName $FamilyName
    }
    else
    {
        Set-SCVirtualHardDisk -VirtualHardDisk $disk -VirtualizationPlatform $VirtualizationPlatform -OperatingSystem $OperatingSystem -Release $Release -FamilyName $FamilyName -Tag @("WindowsServer2012", "R2") -ProductKey $ProductKey
    }
}

Function Set-OSDisk
{
    [CmdletBinding()]
    Param(
        [Parameter(Mandatory=$true)]
        [string]$ProductKey,

        [Parameter(Mandatory=$true)]
        [string]$SourceOSVHDName,

		[Parameter(Mandatory=$true)]
        [string]$OSDiskRelease,

        [Parameter(Mandatory=$true)]
        [string]$VMMlibraryName
            
    )
    
    if(-not (Test-Path $SourceOSVHDName))
    {
        Log "Source Image VHD ($SourceOSVHDName) does not exist." "Error"
        return
    }
       
		Get-VMMServerConnection -CPSRack $IsCPSRack   # Determine the SCVMM Server Name based on the Rack
	
		$osdisk = Get-Item -Path $SourceOSVHDName
    
		Log "Mount the image" "Info"
		Mount-DiskImage -ImagePath $osdisk.FullName
		$drive = Get-DiskImage –ImagePath $osdisk.FullName | Get-Disk | Get-Partition | Get-Volume
    
		Log "Copy DSC resources into OS image" "Info"
		Copy-Item .\DSCResources\* ($drive.DriveLetter + ":\Program Files\WindowsPowerShell\Modules") -Recurse -Force
    
		Log "Dismount OS image" "Info"
		Dismount-DiskImage -ImagePath $osdisk.FullName

		Log "Copy OS image to Library Share" "Info"
		$libraryShare = Get-SCLibraryShare | ? name -eq $VMMlibraryName
		Copy-Item $osdisk.FullName $libraryShare.Path -Force

		Log "Refresh Library Share" "Info"
		$libraryShare | Read-SCLibraryShare 

		Log "Set OS Disk metadata" "Info"
		$release = $OSDiskRelease
		$defaultrelease = "1.0.0.0"
		$familyName = "Windows Server 2012 R2 Datacenter"

		Get-SCVirtualHardDisk -Release $release -FamilyName $familyName | Set-SCVirtualHardDisk -Release $defaultrelease
        
		Set-DiskMetadata -DiskSharedPath (Join-Path $libraryShare.Path $osdisk.Name) -VirtualizationPlatform HyperV -OperatingSystem "Windows Server 2012 R2 Datacenter" -Release $release -FamilyName $familyName -ProductKey $ProductKey
	
}

Function Set-DataDisk
{
    [CmdletBinding()]
    Param(
        
        [Parameter(Mandatory=$true)]
        [string]$DataDiskVHDName,

        [Parameter(Mandatory=$true)]
        [string]$FamilyName,

        [Parameter(Mandatory=$true)]
        [string]$Release,

        [Parameter(Mandatory=$true)]
        [string]$VMMlibraryName
    )

    if(-not (Test-Path $DataDiskVHDName))
    {
        Log "Source Image VHD ($DataDiskVHDName) does not exist." "Error"
        return
    }

    $vhdpath = Join-Path $PSScriptRoot $DataDiskVHDName
      
    $prefix = Get-MachineNamePrefix
    Get-SCVMMServer -ComputerName "$prefix-HA-VMM"

    Log "Copy Data disk to Library Share" "Info"
    $libraryShare = Get-SCLibraryShare | ? name -eq $VMMlibraryName
    Copy-Item $vhdpath $libraryShare.Path -Force

    Log "Refresh Library Share" "Info"
    $libraryShare | Read-SCLibraryShare 

    Set-DiskMetadata -DiskSharedPath (Join-Path $libraryShare.Path $DataDiskVHDName) -VirtualizationPlatform HyperV -OperatingSystem "None" -Release $Release -FamilyName $FamilyName
}

ipmo virtualmachinemanager
    
if($UpdateArtifacts)
{
    #Start deployment
    Log "Start to import runbooks" -level Info
    Import-Runbooks
    Import-Runbooks
    Log "Complete to import runbooks" -level Info

    #Start import gallery item

    Log "Start to import galleryitems" -level Info
    Import-GalleryItem
    Log "Complete to galleryitems" -level Info

    Log "Start to set SPF Event Registration" -level Info
    New-SPFEventRegistration
    Log "Complete to set SPF Event Registration" -level Info

    Log "Start to import resource extension" -level Info
    Import-ResourceExtension
    Log "Complete to resource extension" -level Info

    Log "Start to Update Tenant provisioning assets" -level Info

    Update-TenantProvisioningAssets -assetDescriptionType "Environmental"
    Update-TenantProvisioningAssets -assetDescriptionType "Notifications"
    if(-NOT [System.String]::IsNullOrEmpty($Password))
    {
        Update-TenantProvisioningAssets -assetDescriptionType "Passwords" -Password $Password
    }
    Update-TenantProvisioningAssets -assetDescriptionType "VMRoles"
    Update-TenantProvisioningAssets -assetDescriptionType "Disks"

    Log "Complete to Update Tenant provisioning assets" -level Info
}

# Deploy PCS specific Gallery Item and OS Disk
if($UpdatePCSArtifacts -eq $true)
{
        
    #Start import gallery item

    Log "Start to import galleryitems" -level Info
    Import-GalleryItem   # This can be enabled once the spfAdmin is installed on Console
    Log "Complete to galleryitems" -level Info
        
    Log "Start to import resource extension" -level Info
    Import-ResourceExtension
    Log "Complete to resource extension" -level Info

}
# Add check to make sure the VHD did not exist on the share

if(-NOT [System.String]::IsNullOrEmpty($SourceOSVHDName))
{
    Set-OSDisk -ProductKey Y4TGP-NPTV9-HTC2H-7MGQ3-DV4TW -SourceOSVHDName $SourceOSVHDName -OSDiskRelease $OSDiskRelease -VMMlibraryName $VMMlibraryName 
}

if(-NOT [System.String]::IsNullOrEmpty($SQLInstallationDiskVHDName))
{
    Set-DataDisk -DataDiskVHDName $SQLInstallationDiskVHDName -FamilyName "SQLServerDataDisk" -Release "12.0.0.0" -VMMlibraryName $VMMlibraryName
}

if(-NOT [System.String]::IsNullOrEmpty($ExchangeInstallationDiskVHDName))
{
    Set-DataDisk -DataDiskVHDName $ExchangeInstallationDiskVHDName -FamilyName "ExchangeSourceDisk" -Release "1.0.0.0" -VMMlibraryName $VMMlibraryName
}

if(-NOT [System.String]::IsNullOrEmpty($SharePointInstallationDiskVHDName))
{
    Set-DataDisk -DataDiskVHDName $SharePointInstallationDiskVHDName -FamilyName "SharePointUpdatedSource" -Release "2.0.0.0" -VMMlibraryName $VMMlibraryName
}

if(-NOT [System.String]::IsNullOrEmpty($SkypeForBusinessInstallationDiskVHDName))
{
    Set-DataDisk -DataDiskVHDName $SkypeForBusinessInstallationDiskVHDName -FamilyName "SkypeForBusinessSorceDisk" -Release "1.0.0.0" -VMMlibraryName $VMMlibraryName
}

